/* Copyright (c) 2018 vesoft inc. All rights reserved.
 *
 * This source code is licensed under Apache 2.0 License,
 * attached with Common Clause Condition 1.0, found in the LICENSES directory.
 */

#include "service/QueryInstance.h"

#include "common/base/Base.h"
#include "executor/ExecutionError.h"
#include "executor/Executor.h"
#include "optimizer/OptRule.h"
#include "parser/ExplainSentence.h"
#include "planner/ExecutionPlan.h"
#include "planner/PlanNode.h"
#include "scheduler/Scheduler.h"
#include "service/GraphFlags.h"
#include "validator/Validator.h"

using nebula::opt::Optimizer;
using nebula::opt::OptRule;
using nebula::opt::RuleSet;

namespace nebula {
namespace graph {

QueryInstance::QueryInstance(std::unique_ptr<QueryContext> qctx) {
    qctx_ = std::move(qctx);
    scheduler_ = std::make_unique<Scheduler>(qctx_.get());
    std::vector<const RuleSet *> rulesets{&RuleSet::defaultRules()};
    if (FLAGS_enable_optimizer) {
        rulesets.emplace_back(&RuleSet::queryRules());
    }
    optimizer_ = std::make_unique<Optimizer>(qctx_.get(), std::move(rulesets));
}

void QueryInstance::execute() {
    Status status = validateAndOptimize();
    if (!status.ok()) {
        onError(std::move(status));
        return;
    }

    if (!explainOrContinue()) {
        onFinish();
        return;
    }

    scheduler_->schedule()
        .then([this](Status s) {
            if (s.ok()) {
                this->onFinish();
            } else {
                this->onError(std::move(s));
            }
        })
        .onError([this](const ExecutionError &e) { onError(e.status()); })
        .onError([this](const std::exception &e) { onError(Status::Error("%s", e.what())); });
}

Status QueryInstance::validateAndOptimize() {
    auto *rctx = qctx()->rctx();
    VLOG(1) << "Parsing query: " << rctx->query();
    auto result = GQLParser().parse(rctx->query());
    NG_RETURN_IF_ERROR(result);
    sentence_ = std::move(result).value();

    NG_RETURN_IF_ERROR(Validator::validate(sentence_.get(), qctx()));

    auto rootStatus = optimizer_->findBestPlan(qctx_->plan()->root());
    NG_RETURN_IF_ERROR(rootStatus);
    auto newRoot = std::move(rootStatus).value();
    qctx_->setPlan(std::make_unique<ExecutionPlan>(const_cast<PlanNode *>(newRoot)));

    return Status::OK();
}

bool QueryInstance::explainOrContinue() {
    if (sentence_->kind() != Sentence::Kind::kExplain) {
        return true;
    }
    qctx_->fillPlanDescription();
    return static_cast<const ExplainSentence *>(sentence_.get())->isProfile();
}

void QueryInstance::onFinish() {
    auto rctx = qctx()->rctx();
    VLOG(1) << "Finish query: " << rctx->query();
    auto ectx = qctx()->ectx();
    auto latency = rctx->duration().elapsedInUSec();
    rctx->resp().set_latency_in_us(latency);
    auto &spaceName = rctx->session()->spaceName();
    rctx->resp().set_space_name(spaceName);
    auto name = qctx()->plan()->root()->varName();
    if (ectx->exist(name)) {
        auto &&value = ectx->moveValue(name);
        if (value.type() == Value::Type::DATASET) {
            auto result = value.moveDataSet();
            if (!result.colNames.empty()) {
                rctx->resp().set_data(std::move(result));
            } else {
                LOG(ERROR) << "Empty column name list";
                rctx->resp().set_error_code(cpp2::ErrorCode::E_EXECUTION_ERROR);
                rctx->resp().set_error_msg("Internal error: empty column name list");
            }
        }
    }

    if (qctx()->planDescription() != nullptr) {
        rctx->resp().set_plan_desc(std::move(*qctx()->planDescription()));
    }

    rctx->finish();

    // The `QueryInstance' is the root node holding all resources during the execution.
    // When the whole query process is done, it's safe to release this object, as long as
    // no other contexts have chances to access these resources later on,
    // e.g. previously launched uncompleted async sub-tasks, EVEN on failures.
    delete this;
}

void QueryInstance::onError(Status status) {
    LOG(ERROR) << status;
    auto *rctx = qctx()->rctx();
    if (status.isSyntaxError()) {
        rctx->resp().set_error_code(cpp2::ErrorCode::E_SYNTAX_ERROR);
    } else if (status.isStatementEmpty()) {
        rctx->resp().set_error_code(cpp2::ErrorCode::E_STATEMENT_EMTPY);
    } else if (status.isSemanticError()) {
        rctx->resp().set_error_code(cpp2::ErrorCode::E_SEMANTIC_ERROR);
    } else {
        rctx->resp().set_error_code(cpp2::ErrorCode::E_EXECUTION_ERROR);
    }
    auto &spaceName = rctx->session()->spaceName();
    rctx->resp().set_space_name(spaceName);
    rctx->resp().set_error_msg(status.toString());
    auto latency = rctx->duration().elapsedInUSec();
    rctx->resp().set_latency_in_us(latency);
    rctx->finish();
    delete this;
}

}   // namespace graph
}   // namespace nebula
